@@ -9,7 +9,7 @@
 #if defined(__i386__) || defined (__x86_64__) || defined(__ia64__)
 #include <sys/io.h>
 #endif
-#include <sys/pci.h>
+#include <linux/pci.h>
 
 #include "hd.h"
 #include "hd_int.h"
@@ -1044,7 +1044,7 @@ void parse_mpconfig(hd_data_t *hd_data, memory_range_t *mem, smp_info_t *smp)
       hd_log_hex(hd_data, 1, len, mem->data + cfg_len + u);
       ADD2LOG("\n");
       if(len < 2) {
-        ADD2LOG("  oops: invalid record lenght\n");
+        ADD2LOG("  oops: invalid record length\n");
         break;
       }
     }
